__all__ = ("run_moving_fractions",)

import argparse
import glob
import os

import numpy as np

from rubin_sim.utils import survey_start_mjd

from . import batches as batches
from . import db as db
from . import metricBundles as mmb


def run_moving_fractions():
    """Calculate completeness and fractions for moving object metrics."""
    parser = argparse.ArgumentParser(description="Run moving object metrics for a particular opsim run.")
    parser.add_argument(
        "--work_dir",
        type=str,
        default=".",
        help="Output (and input) directory for moving object metrics. Default '.'.",
    )
    parser.add_argument(
        "--metadata",
        type=str,
        default=None,
        help="Select only files matching this metadata string. Default None (all files).",
    )
    parser.add_argument(
        "--h_mark",
        type=float,
        default=None,
        help="H value at which to calculate cumulative/differential completeness, etc."
        "Default (None) will be set to plot_dict value or median of H range.",
    )
    parser.add_argument(
        "--n_years_max",
        type=int,
        default=10,
        help="Maximum number of years out to which to evaluate completeness." "Default 10.",
    )
    parser.add_argument(
        "--start_time",
        type=float,
        default=None,
        help="Time at start of survey (to set time for summary metrics).",
    )
    args = parser.parse_args()

    # Default parameters for metric setup.
    if args.start_time is None:
        start_time = survey_start_mjd()
    else:
        start_time = args.start_time
    stepsize = 365 / 6.0
    times = np.arange(0, args.n_years_max * 365 + stepsize / 2, stepsize)
    times += start_time

    # Create a results Db.
    results_db = db.ResultsDb(out_dir=args.work_dir)

    # Just read in all metrics in the (joint or single) directory, then run completeness and fraction
    # summaries, using the methods in the batches.
    if args.metadata is None:
        matchstring = os.path.join(args.work_dir, "*MOOB.npz")
    else:
        matchstring = os.path.join(args.work_dir, f"*{args.metadata}*MOOB.npz")
    metricfiles = glob.glob(matchstring)
    metric_names = []
    for m in metricfiles:
        mname = os.path.split(m)[-1].replace("_MOOB.npz", "")
        metric_names.append(mname)

    bdict = {}
    for mName, mFile in zip(metric_names, metricfiles):
        bdict[mName] = mmb.create_empty_mo_metric_bundle()
        bdict[mName].read(mFile)

    first = bdict[metric_names[0]]
    figroot = f"{first.run_name}"
    # adding args.metadata to the figroot means that we can narrow down *which* subset of files to work on
    # this lets us add different h_mark values to the output plots, if desired (leaving None/default is ok)
    if args.metadata != ".":
        figroot += f"_{args.metadata}"

    # Calculate completeness. This utility writes these to disk.
    bdict_completeness = batches.run_completeness_summary(
        bdict, args.h_mark, times, args.work_dir, results_db
    )

    # Plot some of the completeness results.
    batches.plot_completeness(
        bdict_completeness,
        figroot=figroot,
        results_db=results_db,
        out_dir=args.work_dir,
    )

    # Calculate fractions of population for characterization. This utility writes these to disk.
    bdict_fractions = batches.run_fraction_summary(bdict, args.h_mark, args.work_dir, results_db)
    # Plot the fractions for colors and lightcurves.
    batches.plot_fractions(bdict_fractions, figroot=figroot, results_db=results_db, out_dir=args.work_dir)

    # Plot nObs and arcLength.
    for k in bdict:
        if "NObs" in k:
            batches.plot_single(bdict[k], results_db=results_db, out_dir=args.work_dir)
        if "ObsArc" in k:
            batches.plot_single(bdict[k], results_db=results_db, out_dir=args.work_dir)

    # Plot the number of chances of discovery metric - this is different than completeness
    # As it plots the metric value directly
    for k in bdict:
        if "DiscoveryNChances" in k and "3_pairs_in_15_nights_detection_loss" in k:
            batches.plot_single(bdict[k], results_db=results_db, out_dir=args.work_dir)
        if "MagicDiscovery" in k:
            batches.plot_single(bdict[k], results_db=results_db, out_dir=args.work_dir)
        if "HighVelocity" in k:
            batches.plot_single(bdict[k], results_db=results_db, out_dir=args.work_dir)

    # Plot likelihood of detecting activity.
    batches.plot_activity(bdict, figroot=figroot, results_db=results_db, out_dir=args.work_dir)



__all__ = ("StringCountMetric",)

from collections import Counter

import numpy as np

from .base_metric import BaseMetric


class Keylookerupper:
    """Helper object to unpack dictionary values as reduceFunction results."""

    def __init__(self, key="blank", name=None):
        self.key = key
        self.__name__ = name

    def __call__(self, indict):
        return np.max(indict[self.key])


class StringCountMetric(BaseMetric):
    """Count up the number of times each string appears in a column.

    Dynamically builds reduce functions for each unique string value, so summary sats can be
    named the same as strings in the simData array without knowing the values of those trings ahead of time.
    """

    def __init__(self, metric_name="stringCountMetric", col="filter", percent=False, **kwargs):
        """
        Parameters
        ----------

        col: str ('filter')
            Column name that has strings to look at
        percent : bool (False)
            Normalize and return results as percents ranther than raw count
        """
        if percent:
            units = "percent"
        else:
            units = "count"
        self.percent = percent
        cols = [col]
        super(StringCountMetric, self).__init__(cols, metric_name, units=units, metric_dtype=object, **kwargs)
        self.col = col

    def run(self, data_slice, slice_point=None):
        counter = Counter(data_slice[self.col])
        # convert to a numpy array
        lables = list(counter.keys())
        # Numpy can't handle empty string as a dtype
        lables = [x if x != "" else "blank" for x in lables]
        metric_value = np.zeros(1, dtype=list(zip(lables, [float] * len(counter.keys()))))
        for key in counter:
            if key == "":
                metric_value["blank"] = counter[key]
            else:
                metric_value[key] = counter[key]
        if self.percent:
            norm = sum(metric_value[0]) / 100.0
            # Not sure I really like having to loop here, but the dtype is inflexible
            for key in metric_value.dtype.names:
                metric_value[key] = metric_value[key] / norm

        # Now to dynamically set up the reduce functions
        for i, key in enumerate(metric_value.dtype.names):
            name = key
            self.reduce_funcs[name] = Keylookerupper(key=key, name=name)
            self.reduce_order[name] = i

        return metric_value



import gzip
import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.phot_utils import Bandpass, PhotometricParameters, Sed

ROOT = os.path.abspath(os.path.dirname(__file__))


class TestSedWavelenLimits(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("always")
        self.wmin = 500
        self.wmax = 1500
        self.bandpasswavelen = np.arange(self.wmin, self.wmax + 0.5, 1)
        self.bandpasssb = np.ones(len(self.bandpasswavelen))
        self.testbandpass = Bandpass(wavelen=self.bandpasswavelen, sb=self.bandpasssb)

    def tearDown(self):
        del self.bandpasswavelen
        del self.bandpasssb
        del self.testbandpass
        del self.wmin
        del self.wmax

    def test_sed_wavelen_range(self):
        """Test setting sed with wavelength range different
        from standard values works properly."""
        sedwavelen = self.bandpasswavelen * 1.0
        sedflambda = np.ones(len(sedwavelen))
        testsed = Sed(wavelen=sedwavelen, flambda=sedflambda, name="TestSed")
        np.testing.assert_equal(testsed.wavelen, sedwavelen)
        np.testing.assert_equal(testsed.flambda, sedflambda)
        self.assertEqual(testsed.name, "TestSed")

    def test_sed_bandpass_match(self):
        """Test errors when bandpass and sed do not completely
        overlap in wavelength range."""
        # Test case where they do match (no error message)
        sedwavelen = np.arange(self.wmin, self.wmax + 0.5, 1)
        sedflambda = np.ones(len(sedwavelen))
        testsed = Sed(wavelen=sedwavelen, flambda=sedflambda)
        print("")
        # Test that no warning is made.
        with warnings.catch_warnings(record=True) as wa:
            w, f = testsed.resample_sed(
                wavelen_match=self.testbandpass.wavelen,
                wavelen=testsed.wavelen,
                flux=testsed.flambda,
            )
            self.assertEqual(len(wa), 0)
        np.testing.assert_equal(w, testsed.wavelen)
        np.testing.assert_equal(f, testsed.flambda)
        # Test that warning is given for non-overlap
        # at either top or bottom end of wavelength range.
        sedwavelen = np.arange(self.wmin, self.wmax - 50, 1)
        sedflambda = np.ones(len(sedwavelen))
        testsed = Sed(wavelen=sedwavelen, flambda=sedflambda)
        with warnings.catch_warnings(record=True) as wa:
            testsed.resample_sed(wavelen_match=self.testbandpass.wavelen)
            self.assertEqual(len(wa), 1)
            self.assertIn("non-overlap", str(wa[-1].message))
        np.testing.assert_equal(testsed.flambda[-1:], np.NaN)
        sedwavelen = np.arange(self.wmin + 50, self.wmax, 1)
        sedflambda = np.ones(len(sedwavelen))
        testsed = Sed(wavelen=sedwavelen, flambda=sedflambda)
        with warnings.catch_warnings(record=True) as wa:
            testsed.resample_sed(wavelen_match=self.testbandpass.wavelen)
            self.assertEqual(len(wa), 1)
            self.assertIn("non-overlap", str(wa[-1].message))
        np.testing.assert_equal(testsed.flambda[0], np.NaN)
        np.testing.assert_equal(testsed.flambda[49], np.NaN)

    def test_sed_mag_errors(self):
        """Test error handling at mag and adu calculation levels of sed."""
        sedwavelen = np.arange(self.wmin + 50, self.wmax, 1)
        sedflambda = np.ones(len(sedwavelen))
        testsed = Sed(wavelen=sedwavelen, flambda=sedflambda)
        # Test handling in calc_mag
        with warnings.catch_warnings(record=True) as w:
            mag = testsed.calc_mag(self.testbandpass)
            self.assertEqual(len(w), 1)
            self.assertIn("non-overlap", str(w[-1].message))
        np.testing.assert_equal(mag, np.NaN)
        # Test handling in calc_adu
        with warnings.catch_warnings(record=True) as w:
            adu = testsed.calc_adu(self.testbandpass, phot_params=PhotometricParameters())
            self.assertEqual(len(w), 1)
            self.assertIn("non-overlap", str(w[-1].message))
        np.testing.assert_equal(adu, np.NaN)
        # Test handling in calc_flux
        with warnings.catch_warnings(record=True) as w:
            flux = testsed.calc_flux(self.testbandpass)
            self.assertEqual(len(w), 1)
            self.assertIn("non-overlap", str(w[-1].message))
        np.testing.assert_equal(flux, np.NaN)


class TestSedName(unittest.TestCase):
    def setUp(self):
        self.wmin = 500
        self.wmax = 1500
        self.wavelen = np.arange(self.wmin, self.wmax + 0.5, 1)
        self.flambda = np.ones(len(self.wavelen))
        self.name = "TestSed"
        self.testsed = Sed(self.wavelen, self.flambda, name=self.name)

    def tearDown(self):
        del self.wmin, self.wmax, self.wavelen, self.flambda
        del self.name
        del self.testsed

    def test_set_name(self):
        self.assertEqual(self.testsed.name, self.name)

    def test_redshift_name(self):
        testsed = Sed(self.testsed.wavelen, self.testsed.flambda, name=self.testsed.name)
        redshift = 0.2
        testsed.redshift_sed(redshift=redshift)
        newname = testsed.name + "_Z" + "%.2f" % (redshift)
        testsed.name = newname
        self.assertEqual(testsed.name, newname)


class SedBasicFunctionsTestCase(unittest.TestCase):
    long_message = True

    def test_read_sed_flambda(self):
        """
        Test how read_sed_flambda handles the reading of SED filenames
        when we fail to correctly specify their gzipped state.
        """

        scratch_dir = tempfile.mkdtemp(prefix="test_read_sed_flambda", dir=ROOT)

        rng = np.random.RandomState(88)
        zipped_name = os.path.join(scratch_dir, "zipped_sed.txt.gz")
        unzipped_name = os.path.join(scratch_dir, "unzipped_sed.txt")
        if os.path.exists(zipped_name):
            os.unlink(zipped_name)
        if os.path.exists(unzipped_name):
            os.unlink(unzipped_name)
        wv = np.arange(100.0, 1000.0, 10.0)
        flux = rng.random_sample(len(wv))
        with gzip.open(zipped_name, "wt") as output_file:
            for ww, ff in zip(wv, flux):
                output_file.write("%e %e\n" % (ww, ff))
        with open(unzipped_name, "wt") as output_file:
            for ww, ff in zip(wv, flux):
                output_file.write("%e %e\n" % (ww, ff))

        ss = Sed()
        ss.read_sed_flambda(zipped_name)
        ss.read_sed_flambda(zipped_name[:-3])
        ss.read_sed_flambda(unzipped_name)
        ss.read_sed_flambda(unzipped_name + ".gz")

        # make sure an error is raised when you try to read
        # a file that does not exist
        with self.assertRaises(IOError) as context:
            ss.read_sed_flambda(os.path.join(scratch_dir, "nonsense.txt"))
        self.assertIn("sed file", context.exception.args[0])

        if os.path.exists(scratch_dir):
            shutil.rmtree(scratch_dir)

    def test_eq(self):
        """
        Test that __eq__ in Sed works correctly
        """
        sed_dir = os.path.join(get_data_dir(), "tests", "cartoonSedTestData", "starSed", "kurucz")
        list_of_seds = os.listdir(sed_dir)
        sedname1 = os.path.join(sed_dir, list_of_seds[0])
        sedname2 = os.path.join(sed_dir, list_of_seds[1])
        ss1 = Sed()
        ss1.read_sed_flambda(sedname1)
        ss2 = Sed()
        ss2.read_sed_flambda(sedname2)
        ss3 = Sed()
        ss3.read_sed_flambda(sedname1)

        self.assertFalse(ss1 == ss2)
        self.assertTrue(ss1 != ss2)
        self.assertTrue(ss1 == ss3)
        self.assertFalse(ss1 != ss3)

        ss3.flambda_tofnu()

        self.assertFalse(ss1 == ss3)
        self.assertTrue(ss1 != ss3)

    def test_cache(self):
        """
        Verify that loading an SED from the cache gives identical
        results to loading the same SED from ASCII (since we are
        not calling cache_LSST_seds(), as soon as we load an SED
        with read_sed_flambda, it should get stored in the
        _global_misc_sed_cache)
        """
        sed_dir = os.path.join(get_data_dir(), "tests", "cartoonSedTestData", "starSed", "kurucz")

        sed_name_list = os.listdir(sed_dir)
        msg = "An SED loaded from the cache is not " "identical to the same SED loaded from disk"
        for ix in range(5):
            full_name = os.path.join(sed_dir, sed_name_list[ix])
            ss_uncache = Sed()
            ss_uncache.read_sed_flambda(full_name)
            ss_cache = Sed()
            ss_cache.read_sed_flambda(full_name)

            self.assertEqual(ss_cache, ss_uncache, msg=msg)

        # test that modifications to an SED don't get pushed
        # to the cache
        full_name = os.path.join(sed_dir, sed_name_list[0])
        ss1 = Sed()
        ss1.read_sed_flambda(full_name)
        ss2 = Sed()
        ss2.read_sed_flambda(full_name)
        ss2.flambda *= 2.0
        ss3 = Sed()
        ss3.read_sed_flambda(full_name)
        msg = "Changes to SED made it into the cache"
        self.assertEqual(ss1, ss3, msg=msg)
        self.assertNotEqual(ss1, ss2, msg=msg)
        self.assertNotEqual(ss2, ss3, msg=msg)

    def test_calc_ergs(self):
        """
        Test that calc_ergs actually calculates the flux of a source in
        ergs/s/cm^2 by running it on black bodies with flat bandpasses
        and comparing to the Stefan-Boltzmann law.
        """

        boltzmann_k = 1.3807e-16  # in ergs/Kelvin
        planck_h = 6.6261e-27  # in cm^2*g/s
        speed_of_light = 2.9979e10  # in cm/s
        stefan_boltzmann_sigma = 5.6705e-5  # in ergs/cm^2/s/Kelvin

        wavelen_arr = np.arange(10.0, 200000.0, 10.0)  # in nm
        bp = Bandpass(wavelen=wavelen_arr, sb=np.ones(len(wavelen_arr)))

        log10_bb_factor = np.log10(2.0) + np.log10(planck_h)
        log10_bb_factor += 2.0 * np.log10(speed_of_light)
        log10_bb_factor -= 5.0 * (np.log10(wavelen_arr) - 7.0)  # convert wavelen to cm

        for temp in np.arange(5000.0, 7000.0, 250.0):
            log10_exp_arg = np.log10(planck_h) + np.log10(speed_of_light)
            log10_exp_arg -= np.log10(wavelen_arr) - 7.0
            log10_exp_arg -= np.log10(boltzmann_k) + np.log10(temp)

            exp_arg = np.power(10.0, log10_exp_arg)
            log10_bose_factor = -1.0 * np.log10(np.exp(exp_arg) - 1.0)

            # the -7.0 below is because, otherwise, flambda will be in
            # ergs/s/cm^2/cm and we want ergs/s/cm^2/nm
            #
            # the np.pi comes from the integral in the 'Stefan-Boltzmann'
            # section of
            # https://en.wikipedia.org/wiki/Planck%27s_law#Stefan.E2.80.93Boltzmann_law
            #
            bb_flambda = np.pi * np.power(10.0, log10_bb_factor + log10_bose_factor - 7.0)

            sed = Sed(wavelen=wavelen_arr, flambda=bb_flambda)
            ergs = sed.calc_ergs(bp)

            log10_ergs = np.log10(stefan_boltzmann_sigma) + 4.0 * np.log10(temp)
            ergs_truth = np.power(10.0, log10_ergs)

            msg = "\ntemp:%e\nergs: %e\nergs_truth: %e" % (temp, ergs, ergs_truth)
            self.assertAlmostEqual(ergs / ergs_truth, 1.0, 3, msg=msg)

        # Now test it on a bandpass with throughput=0.25 and an wavelength
        # array that is not the same as the SED

        wavelen_arr = np.arange(10.0, 100000.0, 146.0)  # in nm
        bp = Bandpass(wavelen=wavelen_arr, sb=0.25 * np.ones(len(wavelen_arr)))

        wavelen_arr = np.arange(5.0, 200000.0, 17.0)

        log10_bb_factor = np.log10(2.0) + np.log10(planck_h)
        log10_bb_factor += 2.0 * np.log10(speed_of_light)
        log10_bb_factor -= 5.0 * (np.log10(wavelen_arr) - 7.0)  # convert wavelen to cm

        for temp in np.arange(5000.0, 7000.0, 250.0):
            log10_exp_arg = np.log10(planck_h) + np.log10(speed_of_light)
            log10_exp_arg -= np.log10(wavelen_arr) - 7.0
            log10_exp_arg -= np.log10(boltzmann_k) + np.log10(temp)

            exp_arg = np.power(10.0, log10_exp_arg)
            log10_bose_factor = -1.0 * np.log10(np.exp(exp_arg) - 1.0)

            # the -7.0 below is because, otherwise, flambda will be in
            # ergs/s/cm^2/cm and we want ergs/s/cm^2/nm
            #
            # the np.pi comes from the integral in the 'Stefan-Boltzmann'
            # section of
            # https://en.wikipedia.org/wiki/Planck%27s_law#Stefan.E2.80.93Boltzmann_law
            #
            bb_flambda = np.pi * np.power(10.0, log10_bb_factor + log10_bose_factor - 7.0)

            sed = Sed(wavelen=wavelen_arr, flambda=bb_flambda)
            ergs = sed.calc_ergs(bp)

            log10_ergs = np.log10(stefan_boltzmann_sigma) + 4.0 * np.log10(temp)
            ergs_truth = np.power(10.0, log10_ergs)

            msg = "\ntemp: %e\nergs: %e\nergs_truth: %e" % (temp, ergs, ergs_truth)
            self.assertAlmostEqual(ergs / ergs_truth, 0.25, 3, msg=msg)

    def test_mags_vs_flux(self):
        """
        Verify that the relationship between Sed.calc_mag() and Sed.calc_flux()
        is as expected
        """
        wavelen = np.arange(100.0, 1500.0, 1.0)
        flambda = np.exp(-0.5 * np.power((wavelen - 500.0) / 100.0, 2))
        sb = (wavelen - 100.0) / 1400.0

        ss = Sed(wavelen=wavelen, flambda=flambda)
        bp = Bandpass(wavelen=wavelen, sb=sb)

        mag = ss.calc_mag(bp)
        flux = ss.calc_flux(bp)

        self.assertAlmostEqual(ss.mag_from_flux(flux) / mag, 1.0, 10)
        self.assertAlmostEqual(ss.flux_from_mag(mag) / flux, 1.0, 10)


if __name__ == "__main__":
    unittest.main()



__all__ = ("HourglassMetric",)

import numpy as np
from astroplan import Observer
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, get_body, get_sun
from astropy.time import Time
from rubin_scheduler.utils import Site

from .base_metric import BaseMetric


def nearest_val(A, val):
    return A[np.argmin(np.abs(np.array(A) - val))]


class HourglassMetric(BaseMetric):
    """Plot the filters used as a function of time. Must be used with the Hourglass Slicer.
    Will totally fail in the arctic circle."""

    def __init__(
        self,
        telescope="LSST",
        mjd_col="observationStartMJD",
        filter_col="filter",
        night_col="night",
        delta_t=60.0,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.filter_col = filter_col
        self.night_col = night_col
        cols = [self.mjd_col, self.filter_col, self.night_col]
        super(HourglassMetric, self).__init__(col=cols, metric_dtype="object", **kwargs)
        self.telescope = Site(name=telescope)
        self.delta_t = delta_t / 60.0 / 24.0
        self.location = EarthLocation(
            lat=self.telescope.latitude_rad * u.rad,
            lon=self.telescope.longitude_rad * u.rad,
            height=self.telescope.height * u.m,
        )
        self.observer = Observer(location=self.location)

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.mjd_col)
        unights, uindx = np.unique(data_slice[self.night_col], return_index=True)

        mjds = np.arange(np.min(data_slice[self.mjd_col]), np.max(data_slice[self.mjd_col]) + 1, 0.5)

        # Define the breakpoints as where either the filter changes OR
        # there's more than a 2 minute gap in observing
        good = np.where(
            (data_slice[self.filter_col] != np.roll(data_slice[self.filter_col], 1))
            | (
                np.abs(np.roll(data_slice[self.mjd_col], 1) - data_slice[self.mjd_col])
                > 120.0 / 3600.0 / 24.0
            )
        )[0]
        good = np.concatenate((good, [0], [len(data_slice[self.filter_col])]))
        good = np.unique(good)
        left = good[:-1]
        right = good[1:] - 1
        good = np.ravel(list(zip(left, right)))

        names = ["mjd", "midnight", "filter"]
        types = ["float", "float", (np.str_, 1)]
        perfilter = np.zeros((good.size), dtype=list(zip(names, types)))
        perfilter["mjd"] = data_slice[self.mjd_col][good]
        perfilter["filter"] = data_slice[self.filter_col][good]

        # brute force compute midnight times for all days between start and enc of data_slice
        times = Time(mjds, format="mjd")
        # let's just find the midnight before and after each of the pre_night MJD values
        m_after = self.observer.midnight(times, "next")
        m_before = self.observer.midnight(times, "previous")
        midnights = np.unique(np.concatenate([m_before.mjd, m_after.mjd]))
        # calculating midnight can return nans? That seems bad.
        midnights = midnights[np.isfinite(midnights)]
        # chop off any repeats. Need to round because observe.midnight values are not repeatable
        m10 = np.round(midnights * 10)
        _temp, indx = np.unique(m10, return_index=True)
        midnights = midnights[indx]
        names = [
            "mjd",
            "midnight",
            "moonPer",
            "twi6_rise",
            "twi6_set",
            "twi12_rise",
            "twi12_set",
            "twi18_rise",
            "twi18_set",
        ]
        types = ["float"] * len(names)
        pernight = np.zeros(len(midnights), dtype=list(zip(names, types)))
        pernight["midnight"] = midnights
        pernight["mjd"] = midnights

        # now for each perfilter, find the closes midnight
        indx = np.searchsorted(midnights, perfilter["mjd"])
        d1 = np.abs(perfilter["mjd"] - midnights[indx - 1])
        indx[np.where(indx >= midnights.size)] -= 1
        d2 = np.abs(perfilter["mjd"] - midnights[indx])

        perfilter["midnight"] = midnights[indx]
        temp_indx = np.where(d1 < d2)
        perfilter["midnight"][temp_indx] = midnights[indx - 1][temp_indx]
        try:
            mtime = Time(pernight["midnight"], format="mjd")
        except:
            import pdb

            pdb.set_trace()
        pernight["twi12_rise"] = self.observer.twilight_morning_nautical(mtime, which="next").mjd
        pernight["twi12_set"] = self.observer.twilight_evening_nautical(mtime, which="previous").mjd

        pernight["twi18_rise"] = self.observer.twilight_morning_astronomical(mtime, which="next").mjd
        pernight["twi18_set"] = self.observer.twilight_evening_astronomical(mtime, which="previous").mjd

        aa = AltAz(location=self.location, obstime=mtime)
        moon_coords = get_body("moon", mtime).transform_to(aa)
        sun_coords = get_sun(mtime).transform_to(aa)
        ang_dist = sun_coords.separation(moon_coords)
        pernight["moonPer"] = ang_dist.deg / 180 * 100

        return {"pernight": pernight, "perfilter": perfilter}



__all__ = ("ModelObservatory",)

import numpy as np
from rubin_scheduler.scheduler.model_observatory import ModelObservatory as OMO
from rubin_scheduler.site_models import Almanac
from rubin_scheduler.utils import _healbin, survey_start_mjd

# Take the model observatory from the scheduler and subclass and expand to include satellite constellations


class ModelObservatory(OMO):
    """A class to generate a realistic telemetry stream for the scheduler

    Parameters
    ----------
    nside : `int`
        The healpix nside resolution
    mjd_start : `float`
        The MJD to start the observatory up at. Uses util to lookup default if None.
    alt_min : `float`
        The minimum altitude to compute models at (degrees).
    lax_dome : `bool`
        Passed to observatory model. If true, allows dome creep.
    cloud_limit : `float`
        The limit to stop taking observations if the cloud model returns something equal or higher
    sim_to_o : `sim_targetoO`
        If one would like to inject simulated ToOs into the telemetry stream.
    seeing_db : `str`
        If one would like to use an alternate seeing database, filename of sqlite file
    park_after : `float`
        Park the telescope after a gap longer than park_after (minutes)
    init_load_length : `int`
        The length of pre-scheduled sky brighntess to load initially (days).
    alt_limit : `float`
        Altitude limit for considering satellite streaks (degrees).
    satellite_dt : `float`
        The time step to use for computing satellite positions (seconds).
    sat_nside : `int`
        The HEALpix nside to use for satellite streak maps.
    constellation : `rubin_sim.satellite_constellations.Constellation`
        The satellite constellation to use.
    """

    def __init__(
        self,
        nside=None,
        mjd_start=None,
        seed=42,
        alt_min=5.0,
        lax_dome=True,
        cloud_limit=0.3,
        sim_to_o=None,
        seeing_db=None,
        park_after=10.0,
        init_load_length=10,
        sat_nside=64,
        satellite_dt=10.0,
        constellation=None,
        alt_limit=20.0,
    ):
        # Add in the new satellite information
        self.alt_limit = np.radians(alt_limit)
        self.satelite_dt = satellite_dt / 3600.0 / 24.0  # Seconds to days
        self.sat_nside = sat_nside
        self.constellation = constellation

        # Need to do a little fiddle with the MJD since self.mjd needs self.night set now.
        self.mjd_start = survey_start_mjd() if mjd_start is None else mjd_start
        self.almanac = Almanac(mjd_start=self.mjd_start)
        self.night = -1

        # Run the rest of the regular __init__ steps
        super().__init__(
            nside=None,
            mjd_start=self.mjd_start,
            seed=42,
            alt_min=5.0,
            lax_dome=True,
            cloud_limit=0.3,
            sim_to_o=None,
            seeing_db=None,
            park_after=10.0,
            init_load_length=10,
        )

    def return_conditions(self):
        """
        Returns
        -------
        rubin_sim.scheduler.features.conditions object
        """

        # Spot to put in satellite streak prediction maps
        self.conditions.satellite_mjds = self.sat_mjds
        self.conditions.satellite_maps = self.satellite_maps

        # Run the regular return conditions
        super().return_conditions()
        # I guess running super() means return statement gets skipped?
        return self.conditions

    @property
    def mjd(self):
        return self._mjd

    @mjd.setter
    def mjd(self, value):
        self._mjd = value
        self.almanac_indx = self.almanac.mjd_indx(value)
        # Update night if needed
        if self.almanac.sunsets["night"][self.almanac_indx] != self.night:
            self.night = self.almanac.sunsets["night"][self.almanac_indx]
            # Update the satellite prediction map for the night
            self._update_satellite_maps()

    def _update_satellite_maps(self):
        """Make the satellite prediction maps for the night

        will set self.sat_mjds and self.satellite_maps that can then be attached to
        a conditions object in self.return_conditions
        """
        sunset = self.almanac.sunsets["sun_n12_setting"][self.almanac_indx]
        sunrise = self.almanac.sunsets["sun_n12_rising"][self.almanac_indx]

        self.sat_mjds = np.arange(sunset, sunrise, self.satelite_dt)

        # Compute RA and decs for when sun is down
        ras, decs, alts, illums = self.constellation.paths_array(self.sat_mjds)

        below_limit = np.where(alts < self.alt_limit)

        weights = np.zeros(ras.shape, dtype=int)
        weights[illums] = 1
        weights[below_limit] = 0

        satellite_maps = []
        for i, mjd in enumerate(self.sat_mjds):
            spot_map = _healbin(
                ras[:, i][illums[:, i]],
                decs[:, i][illums[:, i]],
                weights[:, i][illums[:, i]],
                self.sat_nside,
                reduce_func=np.sum,
                dtype=int,
                fill_val=0,
            )

            satellite_maps.append(spot_map)

        self.satellite_maps = np.vstack(satellite_maps)



__all__ = ("MafTracking",)

import os
from collections import OrderedDict

import numpy as np

from rubin_sim.maf.utils import get_sim_data

from .maf_run_results import MafRunResults


class MafTracking:
    """
    Class to read MAF's tracking SQLite database (tracking a set of MAF runs)
    and handle the output for web display.
    """

    def __init__(self, database=None):
        """
        Instantiate the (multi-run) layout visualization class.

        Parameters
        ----------
        database :str
           Path to the sqlite tracking database file.
           If not set, looks for 'trackingDb_sqlite.db' file in current directory.
        """
        if database is None:
            database = os.path.join(os.getcwd(), "trackingDb_sqlite.db")

        # Read in the results database.
        cols = [
            "maf_run_id",
            "run_name",
            "run_group",
            "maf_comment",
            "run_comment",
            "db_file",
            "maf_dir",
            "run_version",
            "run_date",
            "maf_version",
            "maf_date",
        ]
        self.runs = get_sim_data(database, "", cols, table_name="runs")
        self.runs = self.sort_runs(self.runs, order=["maf_run_id", "run_name", "maf_comment"])
        self.runs_page = {}

    def run_info(self, run):
        """
        Provide the tracking database information relevant for a given run in a format
        that the jinja2 templates can use.

        Parameters
        ----------
        run : `numpy.NDarray`
           One line from self.runs

        Returns
        -------
        run_info : `OrderedDict`
            Ordered dict version of the numpy structured array.
        """
        runInfo = OrderedDict()
        runInfo["Run Name"] = run["run_name"]
        runInfo["Group"] = run["run_group"]
        runInfo["Maf Comment"] = run["maf_comment"]
        runInfo["Run Comment"] = run["run_comment"]
        runInfo["SQLite File"] = [
            os.path.relpath(run["db_file"]),
            os.path.split(run["db_file"])[1],
        ]
        runInfo["ResultsDb"] = os.path.relpath(os.path.join(run["maf_dir"], "resultsDb_sqlite.db"))
        runInfo["maf_dir"] = run["maf_dir"]
        runInfo["sched_version"] = run["run_version"]
        runInfo["sched_date"] = run["run_date"]
        runInfo["maf_version"] = run["maf_version"]
        runInfo["maf_date"] = run["maf_date"]
        return runInfo

    def sort_runs(self, runs, order=["run_name", "maf_comment", "maf_run_id"]):
        """
        Sort the numpy array of run data.

        Parameters
        ----------
        runs : `numpy.NDarray`
           The runs from self.runs to sort.
        order : `list`
           The fields to use to sort the runs array.

        Returns
        -------
        runs : `numpy.NDarray`
           A sorted numpy array.
        """
        return np.sort(runs, order=order)

    def get_run(self, maf_run_id):
        """
        Set up a mafRunResults object to read and handle the data from an individual run.
        Caches the mafRunResults object, meaning the metric information from a particular run
        is only read once from disk.

        Parameters
        ----------
        maf_run_id : `int`
           maf_run_id value in the tracking database corresponding to a particular MAF run.

        Returns
        -------
        runPage : `MafRunResults`
           A MafRunResults object containing the information about a particular run.
           Stored internally in self.runs_page dict, but also passed back to the tornado server.
        """
        if not isinstance(maf_run_id, int):
            if isinstance(maf_run_id, dict):
                maf_run_id = int(maf_run_id["maf_run_id"][0][0])
            if isinstance(maf_run_id, list):
                maf_run_id = int(maf_run_id[0])
        if maf_run_id in self.runs_page:
            return self.runs_page[maf_run_id]
        match = self.runs["maf_run_id"] == maf_run_id
        maf_dir = self.runs[match]["maf_dir"][0]
        run_name = self.runs[match]["run_name"][0]
        if run_name == "NULL":
            run_name = None
        self.runs_page[maf_run_id] = MafRunResults(maf_dir, run_name)
        return self.runs_page[maf_run_id]



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestVisitGroupsMetric(unittest.TestCase):
    def test_pair_fraction_metric(self):
        metric = metrics.PairFractionMetric(mjd_col="mjd")
        times = np.arange(0, 200, 30) / 60.0 / 24.0
        data = np.core.records.fromarrays([times], names="mjd")
        # These should all have pairs
        result = metric.run(data)
        self.assertEqual(result, 1.0)

        # These should have none
        times = np.arange(0, 400, 100) / 60.0 / 24.0
        data = np.core.records.fromarrays([times], names="mjd")
        result = metric.run(data)
        self.assertEqual(result, 0.0)

        # If those are right, then this should be 50%
        t1 = np.arange(0, 200, 30) / 60.0 / 24.0
        t2 = np.arange(0, 400, 100) / 60.0 / 24.0 + 1000

        times = np.append(t1, t2)
        data = np.core.records.fromarrays([times], names="mjd")
        result = metric.run(data)
        expected = np.size(t1) / float(np.size(t1) + np.size(t2))
        self.assertEqual(result, expected)

    def test_visit_groups(self):
        """Test visit groups (solar system groups) metric."""
        # Set up some simple test data.
        tmin = 15.0 / 60.0 / 24.0
        tmax = 90.0 / 60.0 / 24.0
        tstart = 49406.00
        # The test data needs to have times for all observations:
        # set up the times in two stages .. the night
        night = np.array(
            [
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                2,
                3,
                3,
                31,
                31,
                32,
                32,
                33,
                33,
                34,
                34,
                34,
                34,
                35,
                35,
                35,
                36,
                36,
                36,
                37,
                37,
                37,
                38,
                38,
                38,
            ],
            "int",
        )
        #   .. and the time within the night.
        expmjd = np.array(
            [
                tstart,
                tstart + tmin / 10.0,
                tstart + tmin + tmin / 2.0,
                tstart + tmin * 2 + tmax,
                tstart + 2 * tmax,
                tstart + 2 * tmax + tmin / 2.0,  # n0
                tstart,
                tstart + tmax,  # n1 .. only 2 obs but should make a pair
                tstart,  # n2 - only 1 obs, should not make a pair
                tstart,
                tstart + tmin,  # n3 .. only 2 obs but should make a pair
                tstart,
                tstart + tmin,
                tstart,
                tstart + tmin,
                tstart,
                tstart + tmin,  # n31/32/33 - a pair on each night
                tstart,
                tstart + tmin / 10.0,
                tstart + tmax * 2,
                tstart + tmax * 2 + tmin / 10.0,  # n34 .. should make no pairs
                tstart,
                tstart + tmin / 10.0,
                tstart + tmax,  # n35 should make 2.5 (too close at start)
                # n36 should make 2.5 pairs (too close at end)
                tstart,
                tstart + tmax,
                tstart + tmax + tmin / 10.0,
                tstart,
                tstart + tmin,
                tstart + tmax,  # n37 - 3 (regular 3)
                tstart,
                tstart + tmax,
                tstart + tmax * 2,  # n38 - 3 (three, but long spacing)
            ],
            "float",
        )
        expmjd = expmjd + night
        testdata = np.core.records.fromarrays([expmjd, night], names=["expmjd", "night"])
        # Set up metric.
        testmetric = metrics.VisitGroupsMetric(
            time_col="expmjd",
            nights_col="night",
            delta_t_min=tmin,
            delta_t_max=tmax,
            min_n_visits=2,
            window=5,
            min_n_nights=3,
        )
        # and set up a copy, with a higher number of min visits per night
        testmetric2 = metrics.VisitGroupsMetric(
            time_col="expmjd",
            nights_col="night",
            delta_t_min=tmin,
            delta_t_max=tmax,
            min_n_visits=3,
            window=5,
            min_n_nights=3,
        )
        # Run metric for expected results.
        metricval = testmetric.run(testdata)
        # These are the expected results, based on the times above.
        expected_nights = np.array([0, 1, 3, 31, 32, 33, 35, 36, 37, 38])
        expected_numvisits = np.array([5.0, 2, 2, 2, 2, 2, 2.5, 2.5, 3, 3])
        np.testing.assert_equal(metricval["visits"], expected_numvisits)
        np.testing.assert_equal(metricval["nights"], expected_nights)
        # Test reduce methods.
        self.assertEqual(testmetric.reduce_median(metricval), np.median(expected_numvisits))
        self.assertEqual(testmetric.reduce_n_nights_with_n_visits(metricval), len(expected_nights))
        self.assertEqual(testmetric2.reduce_n_nights_with_n_visits(metricval), 3)
        self.assertEqual(testmetric.reduce_n_visits_in_window(metricval), 11)
        self.assertEqual(testmetric2.reduce_n_nights_in_window(metricval), 2)
        self.assertEqual(testmetric.reduce_n_lunations(metricval), 2)
        # Test with a longer (but simpler) date range.
        indnight = np.array([0, 1, 2, 3, 4, 5, 31, 32, 33, 34, 61, 62, 63, 121, 122, 123], "int")
        indtimes = np.array([tstart, tstart + tmin, tstart + tmax], "float")
        expmjd = []
        night = []
        for n in indnight:
            for t in indtimes:
                expmjd.append(float(n + t))
                night.append(n)
        expmjd = np.array(expmjd)
        night = np.array(night)
        testdata = np.core.records.fromarrays([expmjd, night], names=["expmjd", "night"])
        metricval = testmetric.run(testdata)
        self.assertEqual(testmetric.reduce_n_lunations(metricval), 4)
        self.assertEqual(testmetric.reduce_max_seq_lunations(metricval), 3)


if __name__ == "__main__":
    unittest.main()



__all__ = ("generate_ss_commands", "generate_ss")

import argparse
import glob
import os
import sqlite3

import numpy as np
import pandas as pd

import rubin_sim.maf.db as db
from rubin_sim.data import get_data_dir


def generate_ss_commands(
    dbfiles=None,
    pops=None,
    split=False,
):
    if dbfiles is None:
        dbfiles = glob.glob("*.db")
        dbfiles.sort()

    data_dir = os.path.join(get_data_dir(), "orbits")
    orbit_files = {
        "granvik_5k": os.path.join(data_dir, "granvik_5k.txt"),
        "granvik_pha_5k": os.path.join(data_dir, "granvik_pha_5k.txt"),
        "vatiras_granvik_10k": os.path.join(data_dir, "vatiras_granvik_10k.txt"),
        "l7_5k": os.path.join(data_dir, "l7_5k.txt"),
        "mba_5k": os.path.join(data_dir, "mba_5k.txt"),
        "occ_rmax5_5k": os.path.join(data_dir, "occ_rmax5_5k.txt"),
        "occ_rmax20_5k": os.path.join(data_dir, "occ_rmax20_5k.txt"),
        "trojan_5k": os.path.join(data_dir, "trojan_5k.txt"),
    }

    pos_dir = os.path.join(get_data_dir(), "orbits_precompute")
    position_files = {
        "granvik_5k": os.path.join(pos_dir, "granvik_5k.npz"),
        "granvik_pha_5k": os.path.join(pos_dir, "granvik_pha_5k.npz"),
        "vatiras_granvik_10k": os.path.join(pos_dir, "vatiras_granvik_10k.npz"),
        "l7_5k": os.path.join(pos_dir, "l7_5k.npz"),
        "mba_5k": os.path.join(pos_dir, "mba_5k.npz"),
        "occ_rmax5_5k": os.path.join(pos_dir, "occ_rmax5_5k.npz"),
        "occ_rmax20_5k": os.path.join(pos_dir, "occ_rmax20_5k.npz"),
        "trojan_5k": os.path.join(pos_dir, "trojan_5k.npz"),
    }

    objtypes = {
        "granvik_5k": "NEO",
        "granvik_pha_5k": "PHA",
        "vatiras_granvik_10k": "Vatira",
        "l7_5k": "TNO",
        "mba_5k": "MBA",
        "occ_rmax5_5k": "OCC_r5",
        "occ_rmax20_5k": "OCC_r20",
        "sdo_5k": "SDO",
        "trojan_5k": "Trojan",
    }

    if pops is None:
        # put in order so longer runtime ones are first
        pops = [
            "trojan_5k",
            "l7_5k",
            "mba_5k",
            "granvik_5k",
            "granvik_pha_5k",
            "vatiras_granvik_10k",
            "occ_rmax5_5k",
            "occ_rmax20_5k",
        ]

    elif pops is not None:
        pp = [p for p in orbit_files.keys() if p == pops]
        if len(pp) == 0:
            print(f"Did not find population {pops} in expected types ({list(orbit_files.keys())}")
        pops = [pops]

    runs = [os.path.split(file)[-1].replace(".db", "") for file in dbfiles]
    runs = [run for run in runs if "tracking" not in run]

    # Find mjd_start for each run:
    mjd_starts = {}
    for filename in dbfiles:
        con = sqlite3.connect(filename)
        value = pd.read_sql("select min(observationStartMJD) from observations;", con)
        con.close()
        mjd_starts[filename] = np.floor(value.values.min())

    if not split:
        output_file = open("ss_script.sh", "w")
        for run, filename in zip(runs, dbfiles):
            out_dir = f"{run}_ss"
            try:
                os.mkdir(out_dir)
            except FileExistsError:
                pass
            # Create the results DB so multiple threads don't try to create it later
            # This isn't used in this script, but needs to exist on-disk.
            results_db = db.ResultsDb(out_dir=out_dir)
        for pop in pops:
            for run, filename in zip(runs, dbfiles):
                objtype = objtypes[pop]

                s1 = (
                    f"make_lsst_obs --simulation_db {filename} --orbit_file {orbit_files[pop]}"
                    f" --positions_file {position_files[pop]}"
                )
                s2 = (
                    f"run_moving_calc --obs_file {run}__{pop}_obs.npz"
                    f" --simulation_db {filename} --orbit_file {orbit_files[pop]}"
                    f" --out_dir {run}_ss"
                    f" --objtype {objtype}"
                    f" --start_time {mjd_starts[filename]}"
                )
                s3 = (
                    f"run_moving_fractions --work_dir {run}_ss"
                    f" --metadata {objtype}"
                    f" --start_time {mjd_starts[filename]}"
                )
                print(s1 + " ; " + s2 + " ; " + s3, file=output_file)
    else:
        for run, filename in zip(runs, dbfiles):
            out_dir = f"{run}_ss"
            try:
                os.mkdir(out_dir)
            except FileExistsError:
                pass
            # Create the results DB so multiple threads don't try to create it later
            results_db = db.ResultsDb(out_dir=out_dir)
            outfile = f"{run}_ss_script.sh"
            if split:
                output_file = open(outfile, "w")
            for pop in pops:
                objtype = objtypes[pop]
                if split:
                    splitfiles = glob.glob(os.path.join(data_dir, "split") + f"/*{pop}*")
                    outfile_split = outfile.replace(".sh", f"_{pop}_split.sh")
                    # If the output split file already exists, remove it (as we append, not write)
                    if os.path.isfile(outfile_split):
                        os.remove(outfile_split)
                    for i, splitfile in enumerate(splitfiles):
                        split = os.path.split(splitfile)[-1]
                        split = split.replace(".des", "").replace(".s3m", "").replace(".txt", "")
                        with open(outfile_split, "a") as wi:
                            s1 = (
                                f"make_lsst_obs --simulation_db {filename} --orbit_file {splitfile}"
                                f" --positions_file {position_files[pop]}"
                                f" --out_dir {out_dir}"
                            )
                            s2 = (
                                f"run_moving_calc --obs_file {out_dir}/{run}__{split}_obs.npz"
                                f" --simulation_db {filename} --orbit_file {orbit_files[pop]}"
                                f" --out_dir {out_dir}/{split}"
                                f" --run_name {run}"
                                f" --objtype {objtype}"
                                f" --start_time {mjd_starts[filename]}"
                            )
                            print(s1 + " ; " + s2, file=wi)
                    s3 = (
                        f"run_moving_join --orbit_file {pop}"
                        f" --base_dir {out_dir}"
                        f" --out_dir {out_dir}/sso"
                    )
                    s4 = (
                        f"run_moving_fractions --work_dir {out_dir}/sso"
                        f" --metadata {objtype}"
                        f" --start_time {mjd_starts[filename]}"
                    )
                    print(
                        f"cat {outfile_split} | parallel -j 10 ; {s3}  ; {s4}",
                        file=output_file,
                    )
    output_file.close()


def generate_ss():
    """Generate solar system processing commands."""

    parser = argparse.ArgumentParser(description="Generate solar system processing commands")
    parser.add_argument("--db", type=str, default=None, help="database to process")
    parser.set_defaults(vatiras=False)
    parser.add_argument("--pop", type=str, default=None, help="identify one population to run")
    parser.add_argument("--start_mjd", type=float, default=None, help="start of the sim")
    parser.add_argument(
        "--split",
        dest="split",
        default=False,
        action="store_true",
        help="Split up population files; rejoin during processing",
    )
    args = parser.parse_args()

    if args.db is None:
        # Just look for any .db files in this directory
        db_files = glob.glob("*.db")
        # But remove trackingDb and results_db if they're there
        try:
            db_files.remove("trackingDb_sqlite.db")
        except ValueError:
            pass
        try:
            db_files.remove("resultsDb_sqlite.db")
        except ValueError:
            pass
    elif isinstance(args.db, str):
        db_files = [args.db]
    else:
        db_files = args.db

    generate_ss_commands(
        split=args.split,
        dbfiles=db_files,
        pops=args.pop,
    )



import glob
import sys

# from rubin_sim.utils import ObservationMetaData
import healpy as hp
import numpy as np

# Modifying createStarDensityMap to use GAIA DR1 catalog

# Use the catsim framework to loop over a healpy map and generate a stellar density map

# Connect to fatboy with: ssh -L 51433:fatboy.phys.washington.edu:1433 gateway.astro.washington.edu
# If non-astro user, use simsuser@gateway.astro.washington.edu

if __name__ == "__main__":
    # Hide imports here so documentation builds
    from rubin_sim.catalogs.db import DBObject
    from rubin_sim.utils import angular_separation, halfSpaceFromRaDec

    # from rubin_sim.catalogs.generation.db import CatalogDBObject
    # Import the bits needed to get the catalog to work
    # from rubin_sim.catUtils.baseCatalogModels import *
    # from rubin_sim.catUtils.exampleCatalogDefinitions import *
    # connect to fatboy
    gaia_db = DBObject(
        database="LSSTCATSIM",
        host="fatboy.phys.washington.edu",
        port=1433,
        driver="mssql+pymssql",
    )

    # get all of the column names for the gaia table in a list
    gaia_columns = gaia_db.get_column_names(tableName="gaia_2016")

    # Set up healpy map and ra, dec centers
    nside = 64

    # Set the min to 15 since we saturate there. CatSim max is 28
    mag_max = 15.2
    bins = np.arange(0.0, mag_max, 0.2)
    star_density = np.zeros((hp.nside2npix(nside), np.size(bins) - 1), dtype=float)
    over_max_mask = np.zeros(hp.nside2npix(nside), dtype=bool)
    lat, ra = hp.pix2ang(nside, np.arange(0, hp.nside2npix(nside)))
    dec = np.pi / 2.0 - lat
    ra = np.degrees(ra)
    dec = np.degrees(dec)

    # Square root of pixel area.
    hpsize_deg = hp.nside2resol(nside, arcmin=True) / 60.0

    # Limit things to a 10 arcmin radius
    hpsize_deg = np.min([10.0 / 60.0, hpsize_deg])

    indx_min = 0

    restore_file = glob.glob("gaiaStarDensity_nside_%i.npz" % (nside))
    if len(restore_file) > 0:
        data = np.load(restore_file[0])
        star_density = data["star_density"].copy()
        indx_min = data["icheck"].copy()
        over_max_mask = data["over_max_mask"].copy()

    print("")
    # Look at a cirular area the same area as the healpix it's centered on.
    bound_length = hpsize_deg / np.pi**0.5
    radius = bound_length

    block_area = hpsize_deg**2  # sq deg

    checksize = 1000
    printsize = 10
    npix = float(hp.nside2npix(nside))

    # If the area has more than this number of objects, flag it as a max
    break_limit = 1e6
    chunk_size = 10000
    for i in np.arange(indx_min, int(npix)):
        last_cp = ""
        # wonder what the units of bound_length are...degrees! And it's a radius
        # The newer interface:
        # obs_metadata = ObservationMetaData(bound_type='circle',
        ##                                   pointing_ra=np.degrees(ra[i]),
        #                                   pointing_dec=np.degrees(dec[i]),
        #                                   bound_length=bound_length, mjd=5700)

        # t = dbobj.getCatalog('ref_catalog_star', obs_metadata=obs_metadata)
        hs = halfSpaceFromRaDec(ra[i], dec[i], radius)
        current_level = 7
        n_bits_off = 2 * (21 - current_level)

        tx_list = hs.findAllTrixels(current_level)

        # actually construct the query
        query = "SELECT ra, dec, phot_g_mean_mag "
        query += "FROM gaia_2016 "
        query += "WHERE "
        for i_pair, pair in enumerate(tx_list):
            min_tx = int(pair[0] << n_bits_off)
            max_tx = int((pair[1] + 1) << n_bits_off)
            query += "(htmid>=%d AND htmid<=%d)" % (min_tx, max_tx)
            if i_pair < len(tx_list) - 1:
                query += " OR "

        dtype = np.dtype([("ra", float), ("dec", float), ("mag", float)])

        results = gaia_db.get_arbitrary_chunk_iterator(query, dtype=dtype, chunk_size=10000)
        result = list(results)[0]

        distances = angular_separation(result["ra"], result["dec"], ra[i], dec[i])  # Degrees
        result = result[np.where(distances < radius)]

        import pdb

        pdb.set_trace()
        # I could think of setting the chunksize to something really large, then only doing one chunk?
        # Or maybe setting up a way to break out of the loop if everything gets really dense?
        temp_hist = np.zeros(np.size(bins) - 1, dtype=float)
        counter = 0
        for chunk in results:
            chunk_hist, bins = np.histogram(chunk[colName], bins)
            temp_hist += chunk_hist
            counter += chunk_size
            if counter >= break_limit:
                over_max_mask[i] = True
                break

        star_density[i] = np.add.accumulate(temp_hist) / block_area

        # Checkpoint
        if (i % checksize == 0) & (i != 0):
            np.savez(
                "gaiaStarDensity_nside_%i.npz" % (nside),
                starDensity=star_density,
                bins=bins,
                icheck=i,
                overMaxMask=over_max_mask,
            )
            last_cp = "Checkpointed at i=%i of %i" % (i, npix)
        if i % printsize == 0:
            sys.stdout.write("\r")
            per_complete = float(i) / npix * 100
            sys.stdout.write(r"%.2f%% complete. " % (per_complete) + last_cp)
            sys.stdout.flush()

    np.savez(
        "gaiaStarDensity_nside_%i.npz" % (nside),
        starDensity=star_density,
        bins=bins,
        overMaxMask=over_max_mask,
    )
    print("")
    print("Completed!")



__all__ = ("ChipVendorMetric",)

import numpy as np

from .base_metric import BaseMetric


class ChipVendorMetric(BaseMetric):
    """
    See what happens if we have chips from different vendors
    """

    def __init__(self, cols=None, **kwargs):
        if cols is None:
            cols = []
        super(ChipVendorMetric, self).__init__(
            col=cols, metric_dtype=float, units="1,2,3:v1,v2,both", **kwargs
        )

    def _chip_names2vendor_id(self, chip_name):
        """
        given a list of chipnames, convert to 1 or 2, representing
        different vendors
        """
        vendors = []
        for chip in chip_name:
            # Parse the chip_name string.
            if int(chip[2]) % 2 == 0:
                vendors.append(1)
            else:
                vendors.append(2)
        return vendors

    def run(self, data_slice, slice_point=None):
        if "chipNames" not in list(slice_point.keys()):
            raise ValueError("No chipname info, need to set use_camera=True with a spatial slicer.")

        uvendor_i_ds = np.unique(self._chip_names2vendor_id(slice_point["chipNames"]))
        if np.size(uvendor_i_ds) == 1:
            result = uvendor_i_ds
        else:
            result = 3
        return result



__all__ = ("read_observations",)

import logging
import os

from rubin_sim.maf.utils import get_sim_data

from .orbits import Orbits


def read_observations(simfile, colmap, constraint=None, dbcols=None):
    """Read the opsim database.

    Parameters
    ----------
    simfile : `str`
        Name (& path) of the opsim database file.
    colmap : `dict`
        colmap dictionary (from rubin_sim.maf.batches.ColMapDict)
    constraint : `str`, optional
        Optional SQL constraint (minus 'where') on the opsim data to read from db.
        Default is None.
    dbcols : `list` of `str`, optional
        List of additional columns to query from the db and add to the output observations.
        Default None.

    Returns
    -------
    np.ndarray, dictionary
        The OpSim data read from the database, and the dictionary mapping the column names to the data.
    """
    if "rotSkyPos" not in colmap:
        colmap["rotSkyPos"] = "rotSkyPos"

    # Set the minimum required columns.
    min_cols = [
        colmap["mjd"],
        colmap["night"],
        colmap["ra"],
        colmap["dec"],
        colmap["filter"],
        colmap["exptime"],
        colmap["seeingGeom"],
        colmap["fiveSigmaDepth"],
    ]
    if dbcols is not None:
        min_cols += dbcols

    more_cols = [
        colmap["rotSkyPos"],
        colmap["seeingEff"],
        "solarElong",
        "observationId",
    ]

    cols = min_cols + more_cols
    cols = list(set(cols))
    logging.info("Querying for columns:\n %s" % (cols))

    # Go ahead and query for all of the observations.
    simdata = get_sim_data(simfile, constraint, cols)
    logging.info("Queried data from opsim %s, fetched %d visits." % (simfile, len(simdata)))
    return simdata



from sys import argv

# from astropy.visualization import astropy_mpl_style
# plt.style.use(astropy_mpl_style)
import numpy as np
from astropy.time import Time, TimeDelta

import rubin_sim.maf.db as db
import rubin_sim.maf.metricBundles as metricBundles
import rubin_sim.maf.slicers as slicers
from rubin_sim.maf.metrics import BaseMetric

from .calc_expected_visits import CalcExpectedVisitsMetric


class CadenceOverVisibilityWindowMetric(BaseMetric):
    """Metric to compare the lightcurve cadence produced by LSST over the visibility window
    for a given position in the sky to the desired cadence.

    This metric determines the number of
    visits to a given field (RA,Dec) performed, including all exposures taken
    with the given set of filters.

    It compares the actual number of visits with the maximum possible visits,
    calculated from the visibility window of the field for the given start and
    end dates, and desired cadence.

    The returned result = ([sum_j (n_visits_actual / n_visits_desired)]/N_filters ) * 100 (%)

    For cadences less than 1 day, this is the sum over all anticipated visits
    per night.  For cadences greater than 1 day, this is calculated as a fraction
    of the anticipated number of visits during batches of nights.
    """

    def __init__(
        self,
        filters,
        cadence,
        start_date,
        end_date,
        metric_name="CadenceOverVisibilityWindowMetric",
        ra_col="fieldRA",
        dec_col="fieldDec",
        exp_col="visitExposureTime",
        n_exp_col="numExposures",
        filter_col="filter",
        obstime_col="observationStartMJD",
        visittime_col="visitTime",
        verbose=False,
        **kwargs,
    ):
        """Arguments:
        filters  list Filterset over which to compute the metric
        cadence  list Cadence desired for each filter in units of decimal hours
                e.g. [ 0.5, 1.0, 1.2 ]
        start_date string Start of observing window YYYY-MM-DD
        end_date string End of observing window YYYY-MM-DD
        """

        self.filters = filters
        self.cadence = cadence
        self.start_date = start_date
        self.end_date = end_date
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.exp_col = exp_col
        self.n_exp_col = n_exp_col
        self.obstime_col = obstime_col
        self.visittime_col = visittime_col
        self.filter_col = filter_col
        self.verbose = verbose

        if len(self.filters) != len(self.cadence):
            raise ValueError(
                "ERROR: The list of filters requested must correspond to the list of required cadences"
            )
            exit()

        columns = [
            self.ra_col,
            self.dec_col,
            self.exp_col,
            self.n_exp_col,
            self.obstime_col,
            self.visittime_col,
            self.filter_col,
        ]

        super(CadenceOverVisibilityWindowMetric, self).__init__(col=columns, metric_name=metric_name)

    def run(self, data_slice, slice_point=None):
        t = np.empty(data_slice.size, dtype=list(zip(["time", "filter"], [float, "|S1"])))
        t["time"] = data_slice[self.obstime_col]

        t_start = Time(self.start_date + " 00:00:00")
        t_end = Time(self.end_date + " 00:00:00")
        n_days = int((t_end - t_start).value)
        dates = np.array([t_start + TimeDelta(i, format="jd", scale=None) for i in range(0, n_days, 1)])

        result = 0.0

        for i, f in enumerate(self.filters):
            if self.verbose:
                print(
                    "Calculating the expected visits in filter "
                    + f
                    + " given required cadence "
                    + str(self.cadence[i])
                )

            # Returns a list of the number of visits per night for each pointing
            pointing = [(data_slice[self.ra_col][0], data_slice[self.dec_col][0])]

            visit = CalcExpectedVisitsMetric(
                pointing,
                self.cadence[i],
                self.start_date,
                self.end_date,
                self.filters[i],
                self.ra_col,
                self.dec_col,
                verbose=self.verbose,
            )

            (n_visits_desired, hrs_visibility) = visit.run(data_slice)

            n_visits_actual = []

            for j, d in enumerate(dates):
                idx = np.where(data_slice[self.filter_col] == f)

                actual_visits_per_filter = data_slice[idx]

                tdx = np.where(
                    actual_visits_per_filter[self.obstime_col].astype(int) == int(d.jd - 2400000.5)
                )

                n_visits_actual.append(float(len(actual_visits_per_filter[tdx])))

            # Case 1: Required cadence is less than 1 day, meaning we
            #         anticipate more than 1 observation per night
            if self.cadence[i] <= 24.0:
                for j, d in enumerate(dates):
                    if n_visits_desired[0][j] > 0:
                        night_efficiency = n_visits_actual[j] / float(n_visits_desired[0][j])

                        result += night_efficiency

                result = result / float(len(dates))

            # Case 2: Required cadence is greater than 1 day, meaning we
            #         expect at least 1 observation within batches of nights
            #         self.cadence[i] long
            else:
                n_nights = int(self.cadence[i] / 24.0)

                for j in range(0, len(dates), n_nights):
                    hrs_available = (np.array(hrs_visibility[0][j : j + n_nights])).sum()

                    n_actual = (np.array(n_visits_actual[j : j + n_nights])).sum()

                    if hrs_available >= 1.0 and n_actual > 1:
                        result += 1.0

                result = result / float(len(dates) / n_nights)

        result = (result / float(len(self.filters))) * 100.0

        if self.verbose:
            print("METRIC RESULT: Observing cadence percentage = " + str(result))

        return result


def compute_metric(params):
    """Function to execute the metric calculation when code is called from
    the commandline"""

    obsdb = db.OpsimDatabase("/home/docmaf/my_repoes/data/baseline2018a.db")
    output_dir = "/home/docmaf/"
    results_db = db.ResultsDb(out_dir=output_dir)

    (propids, proptags) = obsdb.fetchPropInfo()
    survey_where = obsdb.createSQLWhere(params["survey"], proptags)

    obs_params = {"verbose": params["verbose"]}

    metric = CadenceOverVisibilityWindowMetric(
        params["filters"], params["cadence"], params["start_date"], params["end_date"], **obs_params
    )

    slicer = slicers.HealpixSlicer(nside=64)
    sqlconstraint = survey_where
    bundle = metricBundles.MetricBundle(metric, slicer, sqlconstraint)

    bgroup = metricBundles.MetricBundleGroup(
        {0: bundle}, obsdb, outDir="newmetric_test", results_db=results_db
    )
    bgroup.run_all()


if __name__ == "__main__":
    if len(argv) == 1:
        print("Metric requires the following commandline sequence, e.g.:")
        print(
            "> python CadenceOverVisibilityWindowMetric.py filters=g,r,i,z cadence=168.0,168.0,1.0,168.0 start_date=2020-01-02 end_date=2020-04-02 survey=option"
        )
        print("  where:")
        print("  filters may be specified as a comma-separated list without spaces")
        print(
            "  cadence is the cadence corresponding to each filter in hours, in a comma-separated list without spaces"
        )
        print("  start_date, end_date are the UTC dates of the start and end of the observing window")
        print("  survey indicates which survey to select data from.  Options are {WFD, DD, NES}")

    else:
        params = {"verbose": False}

        for arg in argv:
            try:
                (key, value) = arg.split("=")

                if key == "filters":
                    params[key] = value.split(",")

                if key == "cadence":
                    cadence_list = []

                    for val in value.split(","):
                        cadence_list.append(float(val))

                    params[key] = cadence_list

                if key in ["start_date", "end_date", "survey"]:
                    params[key] = value

            except ValueError:
                pass

            if "verbose" in arg:
                params["verbose"] = True

        compute_metric(params)



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Provides the coordinate conversions between equatorial and galactic coordinates, as well as to galactic cylindrical coordinates. Two different functions are present that do the conversion, and a third that uses ephem package, for redundancy purposes. For use with Field Star Count metric
import math
import sys

import numpy as np
from scipy.optimize import fsolve

rad1 = np.radians(282.25)
rad2 = np.radians(62.6)
rad3 = np.radians(33.0)


def eq_gal(eq_ra, eq_dec):
    d = np.radians(eq_dec)
    a = np.radians(eq_ra)

    def equations(p):
        b, l, x = p
        f1a = np.cos(d) * (np.cos(a - rad1))
        f2a = np.sin(d) * np.sin(rad2) + np.cos(d) * np.sin(a - rad1) * np.cos(rad2)
        f3a = np.sin(d) * np.cos(rad2) - np.cos(d) * np.sin(a - rad1) * np.sin(rad2)
        f1 = np.cos(b) * np.cos(l - rad3) - f1a
        f2 = np.cos(b) * np.sin(l - rad3) - f2a
        f3 = np.sin(b) - f3a
        return (f1, f2, f3)

    b, l, x = fsolve(equations, (0, 0, 0))
    b_deg = np.degrees(b) % 360  # galactic latitude
    if b_deg >= 270:
        b_deg = b_deg - 360
    if b_deg > 90:
        b_deg = 180 - b_deg
        l = l + np.pi
    l_deg = np.degrees(l) % 360  # galactic longitude
    return b_deg, l_deg
    # http://scienceworld.wolfram.com/astronomy/GalacticCoordinates.html


def eq_gal2(eq_ra, eq_dec):
    d = np.radians(eq_dec)
    p = np.radians(eq_ra)
    AC = np.radians(90.0) - d
    AB = np.radians(62.8717)
    CAB = np.radians(192.8585) - p
    cos_bc = np.sin(d) * np.cos(AB) + np.cos(d) * np.sin(AB) * np.cos(CAB)
    BC = np.arccos(cos_bc)
    AD = np.radians(118.9362)
    CAD = np.radians(266.4051) - p
    cos_cd = np.sin(d) * np.cos(AD) + np.cos(d) * np.sin(AD) * np.cos(CAD)
    cos_cbd = cos_cd / np.sin(BC)
    if cos_cbd > 1:
        cos_cbd = 1
    elif cos_cbd < -1:
        cos_cbd = -1
    CBD = np.arccos(cos_cbd)
    b_deg = 90.0 - np.degrees(BC)
    ad = np.radians(90.0)
    cad = np.radians(282.8595) - p
    coscd = np.cos(cad) * np.cos(d)
    coscbd = np.cos(cad) * np.cos(d) / np.sin(BC)
    if coscbd > 1:
        coscbd = 1
    elif coscbd < -1:
        coscbd = -1
    cbd = np.arccos(coscbd)
    if cbd - CBD < 32.9319:
        l_deg = 360.0 - np.degrees(CBD)
    else:
        l_deg = np.degrees(CBD)
    return b_deg, l_deg


def eq_gal3(eq_ra, eq_dec):
    coordset = ephem.Equatorial(np.radians(eq_ra), np.radians(eq_dec), epoch="2000")
    g = ephem.Galactic(coordset)
    templon, templat = float(g.lon), float(g.lat)
    l_deg = np.degrees(templon)
    b_deg = np.degrees(templat)
    return b_deg, l_deg


def gal_cyn(b_deg, l_deg, dist):
    b_rad = np.radians(b_deg)
    l_rad = np.radians(l_deg)
    Z = np.sin(b_rad) * dist
    xy = np.cos(b_rad) * dist
    x = np.cos(l_rad) * xy
    y = np.sin(l_rad) * xy
    x_new = 8000.0 - x
    R = np.power(x_new**2 + y**2, 0.5)
    rho = np.arctan(y / x)
    return R, rho, Z


if __name__ == "__main__":
    gal_lat, gal_lon = eq_gal2(float(sys.argv[1]), float(sys.argv[2]))
    print(gal_lat, gal_lon)
    R, rho, z = gal_cyn(gal_lat, gal_lon, float(sys.argv[3]))
    print(R, rho, z)



__all__ = ("NEODistStacker",)

import numpy as np

from .base_stacker import BaseStacker
from .general_stackers import FiveSigmaStacker


class NEODistStacker(BaseStacker):
    """
    For each observation, find the max distance to a ~144 km NEO,
    also stack on the x,y position of the object.
    """

    cols_added = ["MaxGeoDist", "NEOHelioX", "NEOHelioY"]

    def __init__(
        self,
        stepsize=0.001,
        max_dist=3.0,
        min_dist=0.3,
        H=22,
        elong_col="solarElong",
        filter_col="filter",
        sun_az_col="sunAz",
        az_col="azimuth",
        m5_col="fiveSigmaDepth",
    ):
        """
        stepsize:  The stepsize to use when solving (in AU)
        max_dist: How far out to try and measure (in AU)
        H: Asteroid magnitude

        Adds columns:
        MaxGeoDist:  Geocentric distance to the NEO
        NEOHelioX: Heliocentric X (with Earth at x,y,z (0,1,0))
        NEOHelioY: Heliocentric Y (with Earth at (0,1,0))

        Note that both opsim v3 and v4 report solarElongation in degrees.
        """
        self.units = ["AU", "AU", "AU"]
        # Also grab things needed for the HA stacker
        self.cols_req = [elong_col, filter_col, sun_az_col, az_col, m5_col]

        self.sun_az_col = sun_az_col
        self.elong_col = elong_col
        self.filter_col = filter_col
        self.az_col = az_col
        self.m5_col = m5_col

        self.H = H
        # Magic numbers (Ivezic '15, private comm.)that convert an asteroid
        # V-band magnitude to LSST filters:
        # V_5 = m_5 + (adjust value)
        self.limiting_adjust = {
            "u": -2.1,
            "g": -0.5,
            "r": 0.2,
            "i": 0.4,
            "z": 0.6,
            "y": 0.6,
        }
        self.deltas = np.arange(min_dist, max_dist + stepsize, stepsize)
        self.G = 0.15

        # Magic numbers from  http://adsabs.harvard.edu/abs/2002AJ....124.1776J
        self.a1 = 3.33
        self.b1 = 0.63
        self.a2 = 1.87
        self.b2 = 1.22

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # This is a pretty rare stacker. Assume we need to rerun
            pass
        elong_rad = np.radians(sim_data[self.elong_col])
        v5 = np.zeros(sim_data.size, dtype=float) + sim_data[self.m5_col]
        for filter_name in self.limiting_adjust:
            fmatch = np.where(sim_data[self.filter_col] == filter_name)
            v5[fmatch] += self.limiting_adjust[filter_name]
        for i, elong in enumerate(elong_rad):
            # Law of cosines:
            # Heliocentric Radius of the object
            R = np.sqrt(1.0 + self.deltas**2 - 2.0 * self.deltas * np.cos(elong))
            # Angle between sun and earth as seen by NEO
            alphas = np.arccos((1.0 - R**2 - self.deltas**2) / (-2.0 * self.deltas * R))
            ta2 = np.tan(alphas / 2.0)
            phi1 = np.exp(-self.a1 * ta2**self.b1)
            phi2 = np.exp(-self.a2 * ta2**self.b2)

            alpha_term = 2.5 * np.log10((1.0 - self.G) * phi1 + self.G * phi2)
            appmag = self.H + 5.0 * np.log10(R * self.deltas) - alpha_term
            # There can be some local minima/maxima when solving, so
            # need to find the *1st* spot where it is too faint, not the
            # last spot it is bright enough.
            too_faint = np.where(appmag > v5[i])

            # Check that there is a minimum
            if np.size(too_faint[0]) == 0:
                sim_data["MaxGeoDist"][i] = 0
            else:
                sim_data["MaxGeoDist"][i] = np.min(self.deltas[too_faint])

        # Make coords in heliocentric
        interior = np.where(elong_rad <= np.pi / 2.0)
        outer = np.where(elong_rad > np.pi / 2.0)
        sim_data["NEOHelioX"][interior] = sim_data["MaxGeoDist"][interior] * np.sin(elong_rad[interior])
        sim_data["NEOHelioY"][interior] = (
            -sim_data["MaxGeoDist"][interior] * np.cos(elong_rad[interior]) + 1.0
        )

        sim_data["NEOHelioX"][outer] = sim_data["MaxGeoDist"][outer] * np.sin(np.pi - elong_rad[outer])
        sim_data["NEOHelioY"][outer] = sim_data["MaxGeoDist"][outer] * np.cos(np.pi - elong_rad[outer]) + 1.0

        # Flip the X coord if sun az is negative?
        if sim_data[self.az_col].min() < -np.pi / 2.0:
            halfval = 180.0
        else:
            halfval = np.pi
        flip = np.where(
            ((sim_data[self.sun_az_col] > halfval) & (sim_data[self.az_col] > halfval))
            | ((sim_data[self.sun_az_col] < halfval) & (sim_data[self.az_col] > halfval))
        )

        sim_data["NEOHelioX"][flip] *= -1.0

        return sim_data



__all__ = ("M5OptimalStacker", "generate_sky_slopes")

import numpy as np
from rubin_scheduler.utils import Site

from .base_stacker import BaseStacker


def generate_sky_slopes():
    """
    Fit a line to how the sky brightness changes with airmass.
    """
    import healpy as hp

    import rubin_sim.skybrightness as sb

    sm = sb.SkyModel(mags=True, moon=False, twilight=False, zodiacal=False)
    mjd = 57000
    nside = 32
    airmass_limit = 2.0
    dec, ra = hp.pix2ang(nside, np.arange(hp.nside2npix(nside)))
    dec = np.pi / 2 - dec
    sm.set_ra_dec_mjd(ra, dec, mjd)
    mags = sm.return_mags()
    filters = mags.dtype.names
    filter_slopes = {}
    for filter_name in filters:
        good = np.where((~np.isnan(mags[filter_name])) & (sm.airmass < airmass_limit))
        pf = np.polyfit(sm.airmass[good], mags[filter_name][good], 1)
        filter_slopes[filter_name] = pf[0]
    print(filter_slopes)


class M5OptimalStacker(BaseStacker):
    """
    Make a new m5 column as if observations were taken on the meridian.
    If the moon is up, assume sky brightness stays the same.

    Assumes seeing scales as airmass^0.6. Uses linear fits for sky and airmass relation.

    Parameters
    ----------
    airmass_col : str ('airmass')
        Column name for the airmass per pointing.
    dec_col : str ('dec_rad')
        Column name for the pointing declination.
    sky_bright_col: str ('filtSkyBrightness')
        Column name for the sky brighntess per pointing.
    filter_col : str ('filter')
        Column name for the filter name.
    m5_col : str ('fiveSigmaDepth')
        Colum name for the five sigma limiting depth per pointing.
    moon_alt_col : str ('moonAlt')
        Column name for the moon altitude per pointing.
    sun_alt_col : str ('sun_alt_col')
        Column name for the sun altitude column.
    site : str ('LSST')
        Name of the site.

    Returns
    -------
    numpy.array
        Adds a column to that is approximately what the five-sigma depth would have
        been if the observation had been taken on the meridian.
    """

    cols_added = ["m5_optimal"]

    def __init__(
        self,
        airmass_col="airmass",
        dec_col="fieldDec",
        sky_bright_col="skyBrightness",
        seeing_col="seeingFwhmEff",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        moon_alt_col="moonAlt",
        sun_alt_col="sunAlt",
        site="LSST",
    ):
        self.site = Site(site)
        self.units = ["mags"]
        self.airmass_col = airmass_col
        self.dec_col = dec_col
        self.sky_bright_col = sky_bright_col
        self.seeing_col = seeing_col
        self.filter_col = filter_col
        self.moon_alt_col = moon_alt_col
        self.sun_alt_col = sun_alt_col
        self.m5_col = m5_col
        self.cols_req = [
            airmass_col,
            dec_col,
            sky_bright_col,
            seeing_col,
            filter_col,
            moon_alt_col,
            sun_alt_col,
        ]
        self.cols_req = list(set(self.cols_req))

    def _run(self, sim_data, cols_present=False):
        # k_atm values from rubin_sim.operations gen_output.py
        k_atm = {"u": 0.50, "g": 0.21, "r": 0.13, "i": 0.10, "z": 0.07, "y": 0.18}
        # Linear fits to sky brightness change, no moon, twilight, or zodiacal components
        # Use generate_sky_slopes to regenerate if needed.
        sky_slopes = {
            "g": -0.52611780327408397,
            "i": -0.67898669252082422,
            "r": -0.61378749766766827,
            "u": -0.27840980367303503,
            "y": -0.69635091524779691,
            "z": -0.69652846002009128,
        }
        min_z_possible = np.abs(np.radians(sim_data[self.dec_col]) - self.site.latitude_rad)
        min_airmass_possible = 1.0 / np.cos(min_z_possible)
        for filter_name in np.unique(sim_data[self.filter_col]):
            delta_sky = sky_slopes[filter_name] * (sim_data[self.airmass_col] - min_airmass_possible)
            delta_sky[
                np.where((sim_data[self.moon_alt_col] > 0) | (sim_data[self.sun_alt_col] > np.radians(-18.0)))
            ] = 0
            # Using Approximation that FWHM~X^0.6. So seeing term in m5 of: 0.25 * log (7.0/FWHMeff )
            # Goes to 0.15 log(FWHM_min / FWHM_eff) in the difference
            m5_optimal = (
                sim_data[self.m5_col]
                - 0.5 * delta_sky
                - 0.15 * np.log10(min_airmass_possible / sim_data[self.airmass_col])
                - k_atm[filter_name] * (min_airmass_possible - sim_data[self.airmass_col])
            )
            good = np.where(sim_data[self.filter_col] == filter_name)
            sim_data["m5_optimal"][good] = m5_optimal[good]
        return sim_data



__all__ = ("SnrWeightedMetric",)

import numpy as np

from rubin_sim.maf.utils import m52snr

from .base_metric import BaseMetric


class SnrWeightedMetric(BaseMetric):
    """Take the SNR weighted average of a column."""

    def __init__(self, col, m5_col="fiveSigmaDepth", metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "SNR Weighted %s" % col
        super(SnrWeightedMetric, self).__init__(col=[m5_col, col], metric_name=metric_name, **kwargs)
        self.m5_col = m5_col
        self.col = col
        self.star_mag = 20.0  # Arbitrary reference, value doesn't matter

    def run(self, data_slice, slice_point=None):
        snr = m52snr(self.star_mag, data_slice[self.m5_col])
        result = np.average(data_slice[self.col], weights=snr)
        return result



__all__ = ("BDParallaxMetric", "VolumeSumMetric")

import healpy as hp
import numpy as np
from scipy import interpolate

import rubin_sim.maf.utils as mafUtils

from .base_metric import BaseMetric


def bd_colors(spec_type):
    """Returns dictionary of potential Brown Dwarf colors

    Parameters
    ----------
    spec_type : `str`
        The desired spetral type of the Brown Dwarf.

    """
    result = {}
    result["L0"] = {"i": 16.00, "z": 14.52, "y": 13.58}
    result["L1"] = {"i": 16.41, "z": 14.93, "y": 13.97}
    result["L2"] = {"i": 16.73, "z": 15.30, "y": 14.33}
    result["L3"] = ({"i": 17.4, "z": 15.88, "y": 14.89},)
    result["L4"] = {"i": 18.35, "z": 16.68, "y": 15.66}
    result["L5"] = {"i": 18.71, "z": 16.94, "y": 15.87}
    result["L6"] = {"i": 19.27, "z": 17.35, "y": 16.27}
    result["L7"] = {"i": 20.09, "z": 18.18, "y": 17.13}
    result["L8"] = {"i": 20.38, "z": 18.10, "y": 17.04}
    result["L9"] = {"i": 20.09, "z": 17.69, "y": 16.57}
    result["T0"] = {"i": 20.22, "z": 17.98, "y": 16.77}
    result["T1"] = {"i": 21.10, "z": 18.84, "y": 17.45}
    result["T2"] = {"i": 21.97, "z": 18.26, "y": 16.75}
    result["T3"] = {"i": 22.50, "z": 18.08, "y": 16.50}
    result["T4"] = {"i": 22.50, "z": 18.02, "y": 16.32}
    result["T5"] = {"i": 22.69, "z": 19.20, "y": 17.43}
    result["T6"] = {"i": 23.00, "z": 19.82, "y": 18.06}
    result["T7"] = {"z": 21.17, "y": 19.34}
    result["T8"] = {"z": 21.52, "y": 19.75}
    result["T9"] = {"z": 21.82, "y": 20.37}
    return result[spec_type]


class BDParallaxMetric(BaseMetric):
    """Calculate the distance to which one could reach a parallax SNR for a given object
    Modification of ParallaxMetric, illustrated in
    https://github.com/jgizis/LSST-BD-Cadence/blob/main/bd_allLT_baseline_17.ipynb

    Uses columns ra_pi_amp and dec_pi_amp, calculated by the ParallaxFactorStacker.

    Parameters
    ----------
    metricName : `str`, opt
        Default 'parallax'.
    m5_col : `str`, opt
        The default column name for m5 information in the input data. Default fiveSigmaDepth.
    filter_col : `str`, opt
        The column name for the filter information. Default filter.
    seeing_col : `str`, opt
        The column name for the seeing information. Since the astrometry errors are based on the physical
        size of the PSF, this should be the FWHM of the physical psf. Default seeingFwhmGeom.
    mags : `dict` (None)
        The absolute magnitude of the obeject in question. Keys of filter name, values in mags.
        Defaults to an L7 spectral type if None.
    distances : np.array
        Distances to try putting the object at (pc).
    atm_err : `float`, opt
        The expected centroiding error due to the atmosphere, in arcseconds. Default 0.01.
    badval : `float`, opt
        The value to return when the metric value cannot be calculated. Default 0.
    """

    def __init__(
        self,
        metric_name="bdParallax",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        seeing_col="seeingFwhmGeom",
        badval=0,
        mags=None,
        parallax_snr=10.0,
        distances=np.arange(10, 200, 10),
        atm_err=0.01,
        normalize=False,
        **kwargs,
    ):
        cols = [m5_col, filter_col, seeing_col, "ra_pi_amp", "dec_pi_amp"]

        units = "pc"
        super().__init__(cols, metric_name=metric_name, units=units, badval=badval, **kwargs)
        # set return types
        self.m5_col = m5_col
        self.seeing_col = seeing_col
        self.filter_col = filter_col
        self.distances = distances
        self.mags = {}
        distance_mod = 5.0 * np.log10(distances) - 5.0
        if mags is None:
            mags = bd_colors("L7")
        for key in mags:
            self.mags[key] = mags[key] + distance_mod
        self.parallax_snr_goal = parallax_snr
        self.atm_err = atm_err
        self.filters = list(self.mags.keys())
        self.parallaxes = 1000.0 / distances  # mas

    def _final_sigma(self, position_errors, ra_pi_amp, dec_pi_amp):
        """Assume parallax in RA and DEC are fit independently, then combined.
        All inputs assumed to be arcsec"""
        sigma_a = position_errors / ra_pi_amp
        sigma_b = position_errors / dec_pi_amp
        sigma_ra = np.sqrt(1.0 / np.sum(1.0 / sigma_a**2, axis=1))
        sigma_dec = np.sqrt(1.0 / np.sum(1.0 / sigma_b**2, axis=1))
        # Combine RA and Dec uncertainties, convert to mas
        sigma = np.sqrt(1.0 / (1.0 / sigma_ra**2 + 1.0 / sigma_dec**2)) * 1e3
        return sigma

    def run(self, data_slice, slice_point=None):
        snr = np.zeros((np.size(self.mags[self.filters[0]]), len(data_slice)), dtype="float")
        # compute SNR for all observations
        for filt in self.filters:
            good = np.where(data_slice[self.filter_col] == filt)[0]
            if np.size(good) > 0:
                snr[:, good] = mafUtils.m52snr(
                    self.mags[str(filt)][:, np.newaxis], data_slice[self.m5_col][good]
                )

        position_errors = mafUtils.astrom_precision(data_slice[self.seeing_col], snr, self.atm_err)
        # uncertainty in the parallax in mas
        sigma = self._final_sigma(position_errors, data_slice["ra_pi_amp"], data_slice["dec_pi_amp"])
        fitted_parallax_snr = self.parallaxes / sigma
        result = self.badval
        # Let's interpolate to the distance where we would get our target SNR
        if np.max(fitted_parallax_snr) >= self.parallax_snr_goal:
            f = interpolate.interp1d(fitted_parallax_snr, self.distances, fill_value="extrapolate")
            result = f(self.parallax_snr_goal)
        return result


class VolumeSumMetric(BaseMetric):
    """Compute the total volume assuming a metric has values of distance"""

    def __init__(self, col=None, metric_name="VolumeSum", nside=None, **kwargs):
        super(VolumeSumMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)
        self.pix_area = hp.nside2pixarea(nside)

    def run(self, data_slice, slice_point=None):
        # volume of sphere, times ratio of pixel area divided by area of sphere
        vols = 1.0 / 3.0 * data_slice[self.colname] ** 3 * self.pix_area
        return np.sum(vols)
